import time

import flask
import sqlalchemy

from coprs import db
from coprs import exceptions
from coprs import forms
from coprs import helpers
from coprs import models

from coprs.views.misc import login_required, page_not_found

from coprs.views.coprs_ns import coprs_ns

from coprs.logic import builds_logic
from coprs.logic import coprs_logic


@coprs_ns.route('/', defaults = {'page': 1})
@coprs_ns.route('/<int:page>/')
def coprs_show(page=1):
    query = coprs_logic.CoprsLogic.get_multiple(flask.g.user, with_mock_chroots=True)
    paginator = helpers.Paginator(query, query.count(), page)

    coprs = paginator.sliced_query
    return flask.render_template('coprs/show.html', coprs=coprs, paginator=paginator)


@coprs_ns.route('/<username>/', defaults = {'page': 1})
@coprs_ns.route('/<username>/<int:page>/')
def coprs_by_owner(username=None, page=1):
    query = coprs_logic.CoprsLogic.get_multiple(flask.g.user,
                                                user_relation='owned',
                                                username=username,
                                                with_mock_chroots=True)
    paginator = helpers.Paginator(query, query.count(), page)

    coprs = paginator.sliced_query
    return flask.render_template('coprs/show.html', coprs=coprs, paginator=paginator)


@coprs_ns.route('/<username>/allowed/', defaults = {'page': 1})
@coprs_ns.route('/<username>/allowed/<int:page>/')
def coprs_by_allowed(username=None, page=1):
    query = coprs_logic.CoprsLogic.get_multiple(flask.g.user,
                                                user_relation='allowed',
                                                username=username,
                                                with_mock_chroots=True)
    paginator = helpers.Paginator(query, query.count(), page)

    coprs = paginator.sliced_query
    return flask.render_template('coprs/show.html', coprs=coprs, paginator=paginator)


@coprs_ns.route('/fulltext/', defaults = {'page': 1})
@coprs_ns.route('/fulltext/<int:page>/')
def coprs_fulltext_search(page=1):
    fulltext = flask.request.args.get('fulltext', '')
    query = coprs_logic.CoprsLogic.get_multiple_fulltext(flask.g.user, fulltext)
    paginator = helpers.Paginator(query, query.count(), page)

    coprs = paginator.sliced_query
    return flask.render_template('coprs/show.html',
                                 coprs=coprs,
                                 paginator=paginator,
                                 fulltext=fulltext)


@coprs_ns.route('/<username>/add/')
@login_required
def copr_add(username):
    form = forms.CoprFormFactory.create_form_cls()()

    return flask.render_template('coprs/add.html', form = form)


@coprs_ns.route('/<username>/new/', methods=['POST'])
@login_required
def copr_new(username):
    """ Receive information from the user on how to create its new copr
    and create it accordingly.
    """
    form = forms.CoprFormFactory.create_form_cls()()
    if form.validate_on_submit():
        copr = coprs_logic.CoprsLogic.add(flask.g.user,
                                          name=form.name.data,
                                          repos=form.repos.data.replace('\n', ' '),
                                          selected_chroots=form.selected_chroots,
                                          description=form.description.data,
                                          instructions=form.instructions.data)
        db.session.commit()
        flask.flash('New copr was successfully created.')

        if form.initial_pkgs.data:
            builds_logic.BuildsLogic.add(flask.g.user,
                                         pkgs=form.initial_pkgs.data.replace('\n', ' '),
                                               copr=copr)
            db.session.commit()
            flask.flash('Initial packages were successfully submitted '
                        'for building.')

        return flask.redirect(flask.url_for('coprs_ns.copr_detail', username=flask.g.user.name, coprname=copr.name))
    else:
        return flask.render_template('coprs/add.html', form = form)


@coprs_ns.route('/<username>/<coprname>/')
def copr_detail(username, coprname):
    query = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname, with_mock_chroots=True)
    form = forms.CoprLegalFlagForm()
    try:
        copr = query.one()
    except sqlalchemy.orm.exc.NoResultFound:
        return page_not_found('Copr with name {0} does not exist.'.format(coprname))

    return flask.render_template('coprs/detail/overview.html',
                                 copr=copr,
                                 form=form)


@coprs_ns.route('/<username>/<coprname>/permissions/')
def copr_permissions(username, coprname):
    query = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname)
    copr = query.first()
    if not copr:
        return page_not_found('Copr with name {0} does not exist.'.format(coprname))

    permissions = coprs_logic.CoprPermissionsLogic.get_for_copr(flask.g.user, copr).all()
    if flask.g.user:
        user_perm = flask.g.user.permissions_for_copr(copr)
    else:
        user_perm = None

    permissions_applier_form = None
    permissions_form = None

    # generate a proper form for displaying
    if flask.g.user:
        if flask.g.user.can_edit(copr):
            permissions_form = forms.PermissionsFormFactory.create_form_cls(permissions)()
        else:
            # https://github.com/ajford/flask-wtf/issues/58
            permissions_applier_form = forms.PermissionsApplierFormFactory.create_form_cls(user_perm)(formdata=None)

    return flask.render_template('coprs/detail/permissions.html',
                                 copr = copr,
                                 permissions_form = permissions_form,
                                 permissions_applier_form = permissions_applier_form,
                                 permissions = permissions,
                                 current_user_permissions = user_perm)


@coprs_ns.route('/<username>/<coprname>/edit/')
@login_required
def copr_edit(username, coprname, form=None):
    query = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname)
    copr = query.first()

    if not copr:
        return page_not_found('Copr with name {0} does not exist.'.format(coprname))
    if not form:
        form = forms.CoprFormFactory.create_form_cls(copr.mock_chroots)(obj=copr)

    return flask.render_template('coprs/detail/edit.html',
                                 copr=copr,
                                 form=form)


@coprs_ns.route('/<username>/<coprname>/update/', methods = ['POST'])
@login_required
def copr_update(username, coprname):
    form = forms.CoprFormFactory.create_form_cls()()
    copr = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname).first()

    if form.validate_on_submit():
        # we don't change owner (yet)
        copr.name = form.name.data
        copr.repos = form.repos.data.replace('\n', ' ')
        copr.description = form.description.data
        copr.instructions = form.instructions.data
        coprs_logic.CoprChrootsLogic.update_from_names(flask.g.user, copr, form.selected_chroots)

        try:
            coprs_logic.CoprsLogic.update(flask.g.user, copr, check_for_duplicates = False) # form validation checks for duplicates
        except (exceptions.ActionInProgressException, exceptions.InsufficientRightsException) as e:
            flask.flash(e)
            db.session.rollback()
        else:
            flask.flash('Copr was updated successfully.')
            db.session.commit()

        return flask.redirect(flask.url_for('coprs_ns.copr_detail', username = username, coprname = copr.name))
    else:
        return copr_edit(username, coprname, form)


@coprs_ns.route('/<username>/<coprname>/permissions_applier_change/', methods = ['POST'])
@login_required
def copr_permissions_applier_change(username, coprname):
    copr = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname).first()
    permission = coprs_logic.CoprPermissionsLogic.get(flask.g.user, copr, flask.g.user).first()
    applier_permissions_form = forms.PermissionsApplierFormFactory.create_form_cls(permission)()

    if not copr:
        return page_not_found('Copr with name {0} does not exist.'.format(name))
    if copr.owner == flask.g.user:
        flask.flash('Owner cannot request permissions for his own copr.')
    elif applier_permissions_form.validate_on_submit():
        # we rely on these to be 0 or 1 from form. TODO: abstract from that
        new_builder = applier_permissions_form.copr_builder.data
        new_admin = applier_permissions_form.copr_admin.data
        coprs_logic.CoprPermissionsLogic.update_permissions_by_applier(flask.g.user, copr, permission, new_builder, new_admin)
        db.session.commit()
        flask.flash('Successfuly updated permissions do Copr "{0}".'.format(copr.name))

    return flask.redirect(flask.url_for('coprs_ns.copr_detail', username = copr.owner.name, coprname = copr.name))


@coprs_ns.route('/<username>/<coprname>/update_permissions/', methods = ['POST'])
@login_required
def copr_update_permissions(username, coprname):
    query = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname)
    copr = query.first()
    permissions = copr.copr_permissions
    permissions_form = forms.PermissionsFormFactory.create_form_cls(permissions)()

    if permissions_form.validate_on_submit():
        # we don't change owner (yet)
        try:
            # if admin is changing his permissions, his must be changed last
            # so that we don't get InsufficientRightsException
            permissions.sort(cmp=lambda x, y: -1 if y.user_id == flask.g.user.id else 1)
            for perm in permissions:
                new_builder = permissions_form['copr_builder_{0}'.format(perm.user_id)].data
                new_admin = permissions_form['copr_admin_{0}'.format(perm.user_id)].data
                coprs_logic.CoprPermissionsLogic.update_permissions(flask.g.user, copr, perm, new_builder, new_admin)
        # for now, we don't check for actions here, as permissions operation don't collide with any actions
        except exceptions.InsufficientRightsException as e:
            db.session.rollback()
            flask.flash(e)
        else:
            db.session.commit()
            flask.flash('Copr permissions were updated successfully.')

    return flask.redirect(flask.url_for('coprs_ns.copr_detail', username = copr.owner.name, coprname = copr.name))


@coprs_ns.route('/<username>/<coprname>/delete/', methods=['GET', 'POST'])
@login_required
def copr_delete(username, coprname):
    form = forms.CoprDeleteForm()
    copr = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname).first()

    if form.validate_on_submit():
        try:
            coprs_logic.CoprsLogic.delete(flask.g.user, copr)
        except (exceptions.ActionInProgressException, exceptions.InsufficientRightsException) as e:
            db.session.rollback()
            flask.flash(e)
            return flask.redirect(flask.url_for('coprs_ns.copr_detail', username=username, coprname=coprname))
        else:
            db.session.commit()
            flask.flash('Copr was deleted successfully.')
            return flask.redirect(flask.url_for('coprs_ns.coprs_by_owner', username=username))
    else:
        return flask.render_template('coprs/detail/delete.html', form=form, copr=copr)


@coprs_ns.route('/<username>/<coprname>/legal_flag/', methods=['POST'])
@login_required
def copr_legal_flag(username, coprname):
    form = forms.CoprLegalFlagForm()
    copr = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname).first()

    legal_flag = models.LegalFlag(raise_message=form.comment.data,
                                  raised_on=int(time.time()),
                                  copr=copr,
                                  reporter=flask.g.user)
    db.session.add(legal_flag)
    db.session.commit()
    flask.flash('Admin was noticed about your report and will investigate the copr shortly.')
    return flask.redirect(flask.url_for('coprs_ns.copr_detail', username=username, coprname=coprname))


@coprs_ns.route('/<username>/<coprname>/repo/')
def generate_repo_file(username, coprname):
    ''' Generate repo file for a given repo name.
        Reponame = username-coprname '''
    # This solution is used because flask splits off the last part after a
    # dash, therefore user-re-po resolves to user-re/po instead of user/re-po
    # FAS usernames may not contain dashes, so this construction is safe.

    reponame = "%s-%s" % (username, coprname)

    if '-' not in reponame:
        return page_not_found('Bad repository name: {0}. Must be username-coprname'.format(reponame))

    username, coprname = reponame.split('-', 1)
    copr = None
    try:
        # query.one() is used since it fetches all builds, unlike query.first().
        copr = coprs_logic.CoprsLogic.get(flask.g.user, username, coprname,
                with_builds=True).one()
    except sqlalchemy.orm.exc.NoResultFound:
        return page_not_found('Copr {0}/{1} does not exist'.format(username, coprname))

    url = ''
    for build in copr.builds:
        if build.results:
            url = build.results
            break

    if not url:
        return page_not_found('Repository not initialized: No finished builds in {0}/{1}.'.format(username, coprname))

    response = flask.make_response(flask.render_template('coprs/copr.repo', copr=copr, url=url))
    response.mimetype='text/plain'
    return response
